/*
 * An Arduino Controlled Digital FM Receiver/Radio with LCD and remote control
 *
 * HARDWARE:
 * Processor board: Arduino Uno
 *
 * Display: 16 char by 2 line LCD display with backlight
 *
 * FM Radio: SparkFun Evaluation Board for Si4703 FM Tuner #WRL-10663
 *
 * Level Converter: SparkFun Logic Level Converter Bi-Directional #BOB-12009
 *
 * IR detection: Radio Shack IR Receiver #2760640 or Adafruit model TSOP38238 ID: 157
 *
 * Remote control: Adafruit 21 key IR remote ID: 389
 *
 * NOTE: A modified version of Si4703_Breakout.h must be used with this code
 *
 * Hardware and software design by: Craig A. Lindley
 * Code compiles on Arduino 1.0 or newer only
 *
 * Revision History
 *
 * Version: 1.0 - 12/08/2013
 *   Initial release
 */

#include <Arduino.h>
#include <avr/eeprom.h>
#include <IRremote.h>
#include <LiquidCrystal.h>
#include <Si4703_Breakout.h>    // NOTE: Must used modified version
#include <Wire.h>

// Default values
#define DEFAULT_CHANNEL 973
#define DEFAULT_VOLUME    1

// LCD messages
#define MSG_WELCOME         "-Welcome to FM-"
#define MSG_MUTE_ON         "Mute On"
#define MSG_MUTE_OFF        "Mute Off"
#define MSG_VOLUME_UP       "Volume Up"
#define MSG_VOLUME_DOWN     "Volume Down"
#define MSG_SEEK_UP         "Seek Up"
#define MSG_SEEK_DOWN       "Seek Down"
#define MSG_PRESET_KEY      "Preset Key 1..9"
#define MSG_PRESET_SELECTED "Preset Selected"

/************************************************************************/
/************************** Hardware Definitions ************************/
/************************************************************************/

// SI-4703 control pins
int SI4703_RESET_PIN  =  2;
int SI4703_SDIO_PIN   = A4;
int SI4703_SCLK_PIN   = A5;

// LCD control pins
int LCD_BACKLIGHT_PIN =  3;    // Pin controlling LCD backlight
int LCD_RS_PIN        =  8;
int LCD_E_PIN         =  9;
int LCD_D0_PIN        = 10;
int LCD_D1_PIN        = 11;
int LCD_D2_PIN        = 12;
int LCD_D3_PIN        = 13;

// Misc control pins
int IR_RECEIVER_PIN   = 4;    // Pin IR receiver is connected to
int POWERON_LED_PIN   = 5;    // Pin controlling power on indicator
int STEREO_LED_PIN    = 6;    // Pin controlling stereo indicator

/************************************************************************/
/***************** IR Receiver Definitions and Functions ****************/
/************************************************************************/

// Declare instance of IR receiver
IRrecv irReceiver(IR_RECEIVER_PIN);

// Result from decode operation
decode_results results;
unsigned long irCode;

// Keycodes for the Car IR remote control from Adafruit
#define CODE_VOLM      0xFD00FF
#define CODE_PLAYPAUSE 0xFD807F
#define CODE_VOLP      0xFD40BF
#define CODE_SETUP     0xFD20DF
#define CODE_UP        0xFDA05F
#define CODE_STOPMODE  0xFD609F
#define CODE_LEFT      0xFD10EF
#define CODE_ENTERSAVE 0xFD906F
#define CODE_RIGHT     0xFD50AF
#define CODE_010P      0xFD30CF
#define CODE_DOWN      0xFDB04F
#define CODE_BACK      0xFD708F
#define CODE_1         0xFD08F7
#define CODE_2         0xFD8877
#define CODE_3         0xFD48B7
#define CODE_4         0xFD28D7
#define CODE_5         0xFDA857
#define CODE_6         0xFD6897
#define CODE_7         0xFD18E7
#define CODE_8         0xFD9867
#define CODE_9         0xFD58A7

// Low level IR code reading function
// Function will return 0 if no IR code available
unsigned long _readIRCode() {

  results.value = 0;

  // Attempt to read an IR code ?
  if (irReceiver.decode(&results)) {

    delay(20);

    // Prepare to receive the next IR code
    irReceiver.resume();
  }
  return results.value;
}

// Read an IR code
// Function will return 0 if no IR code available
unsigned long readIRCode() {

  // Is there an IR code to read ?
  unsigned long code = _readIRCode();
  if (code == 0) {
    // No code so return 0
    return 0;
  }
  // Keep reading until code changes
  while (_readIRCode() == code) {
    ;
  }
  return code;
}

/************************************************************************/
/*********************** LCD Definitions and Functions ******************/
/************************************************************************/

// Declare instance of Liquid Crystal driver
LiquidCrystal lcd(LCD_RS_PIN, LCD_E_PIN, LCD_D0_PIN, LCD_D1_PIN, LCD_D2_PIN, LCD_D3_PIN);

// Clear specified row of LCD
void clearLCDRow(uint8_t row) {

    lcd.setCursor(0, row);
    lcd.print("                ");    
}

/************************************************************************/
/******************** SI-4703 Definitions and Function ******************/
/************************************************************************/

// Declare instance of FM radio
Si4703_Breakout radio(SI4703_RESET_PIN, SI4703_SDIO_PIN, SI4703_SCLK_PIN);

#define POWERCFG    2    // Register
#define DMUTE      14    // Bit within register for muting control
#define STATUSRSSI 10    // Register
#define STEREO      8    // Bit within register for stereo indication

// Polls the SI4703 to see if stereo is being received
void updateStereoIndicator() {

    // Read all of the SI4703 registers
    radio.readRegisters();
    
    // Retrieve the status register
    uint16_t status = radio.si4703_registers[STATUSRSSI];
    
    // If stereo is being received light the indicator otherwise leave it dark
    if (status & (1 << STEREO)) {
        digitalWrite(STEREO_LED_PIN, HIGH);
    }    else    {
        digitalWrite(STEREO_LED_PIN, LOW);
    }
}

// Mute/UnMute the radio
void muteRadio(boolean mute) {
    
    // Read all of the SI4703 registers
    radio.readRegisters();
    
    // Retrieve the status register
    uint16_t powerCfgRegister = radio.si4703_registers[POWERCFG];
    
    if (mute) {
        // Clear the mute bit
        powerCfgRegister &= ~(1 << DMUTE);

    }    else    {

        // Set the mute bit
        powerCfgRegister |= (1 << DMUTE);
    }
    // Update local register's value
    radio.si4703_registers[POWERCFG] = powerCfgRegister;
    
    // Write the updated registers back to the Si4703
    radio.updateRegisters();
}

/************************************************************************/
/********************** EEPROM read/write Functions *********************/
/************************************************************************/

// Read uint8_t from the eeprom @ address
uint8_t eepromReadUint8_t(int address) {

  uint8_t value;

  eeprom_read_block(&value, (void *) address, sizeof(uint8_t));

  return value;
}

// Write uint8_t to the eeprom @ address
void eepromWriteUint8_t(int address, uint8_t value) {

  eeprom_write_block(&value, (void *) address, sizeof(uint8_t));
}

// Read uint16_t from the eeprom @ address
uint16_t eepromReadUint16_t(int address) {

  uint16_t value;

  eeprom_read_block(&value, (void *) address, sizeof(uint16_t));

  return value;
}

// Write uint16_t to the eeprom @ address
void eepromWriteUint16_t(int address, uint16_t value) {

  eeprom_write_block(&value, (void *) address, sizeof(uint16_t));
}

/************************************************************************/
/**************************** Misc Functions ****************************/
/************************************************************************/

// Display the currently tuned channel centered on row 0 of LCD
void displayChannel() {
    
    // Buffers for string formatting
    char buffer1[10];
    char buffer2[16];

    // Clear row 0
    clearLCDRow(0);
    
    // Get channel
    uint16_t channel = radio.getChannel();
    
    // Convert channel to string
    dtostrf(channel / 10.0, 4, 1, buffer1);
    
    // Build complete string
    strcpy(buffer2, "FM - ");
    strcat(buffer2, buffer1);
    strcat(buffer2, " MHz");
    
    uint8_t length = strlen(buffer2);
    int8_t offset = (16 - length ) / 2;
    
    // Position cursor to center channel string
    lcd.setCursor(offset, 0);
    
    // Print the channel string
    lcd.print(buffer2);
}
    
// Display row 1 message
void displayRow1Message(char *message) {
  
    // Clear row 1
    clearLCDRow(1);
    
    uint8_t length = strlen(message);
    int8_t offset = (16 - length ) / 2;
    
    // Position cursor to center message
    lcd.setCursor(offset, 1);
    
    // Print the message
    lcd.print(message);
}

// Retrieve a preset channel. Presets are numbered 1 .. 9
// Preset 10 stores the channel selected when the radio was turned off
// so it can be restored when the radio is turned back on
uint16_t retrievePreset(uint8_t presetNumber) {
    
    return eepromReadUint16_t(presetNumber * 2);        
}

// Store a preset channel
void storePreset(uint8_t presetNumber, uint16_t channel) {
    
    eepromWriteUint16_t(presetNumber * 2, channel);
}

// Process set preset
void processSetPreset(uint16_t channel) {
    
    unsigned long code;    

    while (true) {
        code = readIRCode();                    // Get IR code
        switch (code) {
            case CODE_1:                        // Preset 1
                storePreset(1, channel);
                displayRow1Message(MSG_PRESET_SELECTED);
                return;
                                
            case CODE_2:                        // Preset 2
                storePreset(2, channel);
                displayRow1Message(MSG_PRESET_SELECTED);
                return;
                                
            case CODE_3:                        // Preset 3
                 storePreset(3, channel);
                 displayRow1Message(MSG_PRESET_SELECTED);
                 return;
                                
            case CODE_4:                        // Preset 4
                 storePreset(4, channel);
                 displayRow1Message(MSG_PRESET_SELECTED);
                 return;
                                
            case CODE_5:                        // Preset 5
                 storePreset(5, channel);
                 displayRow1Message(MSG_PRESET_SELECTED);
                 return;
                                
            case CODE_6:                        // Preset 6
                 storePreset(6, channel);
                 displayRow1Message(MSG_PRESET_SELECTED);
                 return;
                                
            case CODE_7:                        // Preset 7
                 storePreset(7, channel);
                 displayRow1Message(MSG_PRESET_SELECTED);
                 return;
                                
            case CODE_8:                        // Preset 8
                 storePreset(8, channel);
                 displayRow1Message(MSG_PRESET_SELECTED);
                 return;
                                
            case CODE_9:                        // Preset 9
                 storePreset(9, channel);
                 displayRow1Message(MSG_PRESET_SELECTED);
                 return;
        }
    }
}

/************************************************************************/
/********** Setup all subsystems before radio starts operation **********/
/************************************************************************/

boolean radioOn;
uint8_t volume;

void setup() {
      
    // Configure control pins as outputs
    pinMode(LCD_BACKLIGHT_PIN, OUTPUT);
    pinMode(POWERON_LED_PIN, OUTPUT);
    pinMode(STEREO_LED_PIN, OUTPUT);

    // Start the IR receiver
    irReceiver.enableIRIn();
  
    // Set number of columns (16) and rows (2) on LCD
    lcd.begin(16, 2);
    
    // Blank display and clear it
    lcd.noDisplay();
    lcd.clear();
    
    // Start the FM receiver
    radio.powerOn();
    
    // Make the radio appear off but tune default channel
    radio.setVolume(0);
    radio.setChannel(DEFAULT_CHANNEL);
    
    radioOn = false;
    volume = DEFAULT_VOLUME;
    
    // Determine if eeprom has been initialized by looking for eeprom signature
    uint8_t sig0 = eepromReadUint8_t(0);
    uint8_t sig1 = eepromReadUint8_t(1);
      
    if ((sig0 != 0xAA) | (sig1 != 0x55)) {
        // EEPROM is uninitialized, so initialize it once
        // First write signature
        eepromWriteUint8_t(0, 0xAA);
        eepromWriteUint8_t(1, 0x55);

        // Then set all presets to 0 meaning not set
        for (int i = 1; i <= 10; i++) {
            storePreset(i, 0);
        }
    }
}

// Program entry point
void loop() { 
    
    // Is an IR key code available ?
    uint32_t keyCode = readIRCode(); 
  
    if (keyCode != 0) {
        // Process remote key code
        switch(keyCode) {

            case CODE_VOLM :                               // Mute the radio
                if (radioOn) {
                    muteRadio(true);
                    displayRow1Message(MSG_MUTE_ON);
                }
                break;
                
            case CODE_VOLP :                               // Un-mute the radio
                if (radioOn) {
                    muteRadio(false);
                    displayRow1Message(MSG_MUTE_OFF);
                }
                break;

            case CODE_PLAYPAUSE:                            // Radio on/off
                if (radioOn) {

                    // Turning radio off
                    radioOn = false;                        // Indicate radio is off
                    storePreset(10, radio.getChannel());          // Store the current channel
                    radio.setVolume(0);                     // Volume to zero
                    lcd.clear();                            // Clear the display
                    lcd.noDisplay();                        // Display off
                    digitalWrite(LCD_BACKLIGHT_PIN, LOW);   // Backlight off
                    digitalWrite(POWERON_LED_PIN, LOW);     // Power indicator off
                    digitalWrite(STEREO_LED_PIN, LOW);      // Stereo indicator off

                }    else    {

                    // Turning radio on
                    digitalWrite(POWERON_LED_PIN, HIGH);     // Power indicator on
                    digitalWrite(LCD_BACKLIGHT_PIN, HIGH);   // Backlight on
                    volume = DEFAULT_VOLUME;
                    radio.setVolume(volume);                 // Volume on
                    muteRadio(false);                        // Unmute radio
                    lcd.display();                           // Display on
                    uint16_t channel = retrievePreset(10);   // See if preset 10 exists
                    if (channel != 0) {
                        radio.setChannel(channel);
                    }
                    displayChannel();                        // Display the tuned channel
                    displayRow1Message(MSG_WELCOME);         // Display welcome message
                    radioOn = true;
                }
                break;
                
                case CODE_UP:                                // Volume up
                    if (radioOn) {
                        volume += 1;                         // Increment volume 
                        if (volume > 15) {                   // Do range check
                            volume = 15;
                        }
                        radio.setVolume(volume);             // Set new volume
                        displayRow1Message(MSG_VOLUME_UP);
                    }
                    break;
                    
                case CODE_DOWN:                              // Volume down
                    if (radioOn) {
                        volume -= 1;                         // Decrement volume 
                        if (volume < 1) {                    // Do range check
                            volume = 1;
                        }
                        radio.setVolume(volume);             // Set new volume
                        displayRow1Message(MSG_VOLUME_DOWN);
                    }
                    break;
                    
                case CODE_LEFT:                              // Seek down
                    if (radioOn) {
                        radio.seekDown();
                        displayChannel();
                        displayRow1Message(MSG_SEEK_DOWN);
                    }
                    break;
                    
                case CODE_RIGHT:                             // Seek up
                    if (radioOn) {
                        radio.seekUp();
                        displayChannel();
                        displayRow1Message(MSG_SEEK_UP);
                    }
                    break;
                    
                case CODE_ENTERSAVE:                           // Set up for setting preset
                    if (radioOn) {
                        displayRow1Message(MSG_PRESET_KEY);    // Setup message
                        uint16_t channel = radio.getChannel(); // Get current channel
                        processSetPreset(channel);             // Process preset
                    }
                    break;
                    
                case CODE_1:            // Preset 1
                    if (radioOn) {
                        uint16_t channel = retrievePreset(1);
                        if (channel != 0) {
                            radio.setChannel(channel);
                            displayChannel();
                            displayRow1Message(MSG_PRESET_SELECTED);
                        }
                    }
                    break;
                        
                case CODE_2:            // Preset 2
                    if (radioOn) {
                        uint16_t channel = retrievePreset(2);
                        if (channel != 0) {
                            radio.setChannel(channel);
                            displayChannel();
                            displayRow1Message(MSG_PRESET_SELECTED);
                        }
                    }
                    break;
                        
                case CODE_3:            // Preset 3
                    if (radioOn) {
                        uint16_t channel = retrievePreset(3);
                        if (channel != 0) {
                            radio.setChannel(channel);
                            displayChannel();
                            displayRow1Message(MSG_PRESET_SELECTED);
                        }
                    }
                    break;
                        
                case CODE_4:            // Preset 4
                    if (radioOn) {
                        uint16_t channel = retrievePreset(4);
                        if (channel != 0) {
                            radio.setChannel(channel);
                            displayChannel();
                            displayRow1Message(MSG_PRESET_SELECTED);
                        }
                    }
                    break;
                        
                case CODE_5:            // Preset 5
                    if (radioOn) {
                        uint16_t channel = retrievePreset(5);
                        if (channel != 0) {
                            radio.setChannel(channel);
                            displayChannel();
                            displayRow1Message(MSG_PRESET_SELECTED);
                        }
                    }
                    break;
                        
                case CODE_6:            // Preset 6
                    if (radioOn) {
                        uint16_t channel = retrievePreset(6);
                        if (channel != 0) {
                            radio.setChannel(channel);
                            displayChannel();
                            displayRow1Message(MSG_PRESET_SELECTED);
                        }
                    }
                    break;
                        
                case CODE_7:            // Preset 7
                    if (radioOn) {
                        uint16_t channel = retrievePreset(7);
                        if (channel != 0) {
                            radio.setChannel(channel);
                            displayChannel();
                            displayRow1Message(MSG_PRESET_SELECTED);
                        }
                    }
                    break;
                        
                case CODE_8:            // Preset 8
                    if (radioOn) {
                        uint16_t channel = retrievePreset(8);
                        if (channel != 0) {
                            radio.setChannel(channel);
                            displayChannel();
                            displayRow1Message(MSG_PRESET_SELECTED);
                        }
                    }
                    break;
                        
                case CODE_9:            // Preset 9
                    if (radioOn) {
                        uint16_t channel = retrievePreset(9);
                        if (channel != 0) {
                            radio.setChannel(channel);
                            displayChannel();
                            displayRow1Message(MSG_PRESET_SELECTED);
                        }
                    }
                    break;
        }
    }    else    {
        // If radio is on stereo indicator can be updated
        // If radio is off, stereo indicator is turned off
        if (radioOn) {
            updateStereoIndicator();            
        }    else    {
            digitalWrite(STEREO_LED_PIN, LOW);      // Stereo indicator off
        }
    }  
}

